# -*- coding: utf-8 -*-
"""
Token based authentication
"""
from urlparse import urlparse

from django.conf import settings
from django.contrib.auth import get_user_model
from requests.exceptions import ConnectionError
from rest_framework.authentication import BaseAuthentication
from .settings import auth_settings
from .models import (
    Token as AuthToken,
    get_or_create_user,
    get_or_create_token
)
from .protocol.cas import cas_validateUser
from .protocol.cas import cas_profile_for_token
from .protocol.globus import (
    globus_profile_for_token, create_user_token_from_globus_profile
)
from .protocol.wso2 import WSO2_JWT

import logging
logger = logging.getLogger(__name__)

User = get_user_model()


def getRequestParams(request):
    """
    Extracts paramters from GET/POST in a Django Request object
    """
    if request.META['REQUEST_METHOD'] == 'GET':
        try:
            # Will only succeed if a GET method with items
            return dict(request.GET.items())
        except:
            pass
    elif request.META['REQUEST_METHOD'] == 'POST':
        try:
            # Will only succeed if a POST method with items
            return dict(request.POST.items())
        except:
            pass
    logger.debug("REQUEST_METHOD is neither GET or POST.")


def getRequestVars(request):
    """
    Extracts parameters from a Django Request object
    Expects ALL or NOTHING. You cannot mix data!
    """
    username = None
    token = None
    api_server = None
    emulate = None
    try:
        # Attempt #1 - SessionStorage - Most reliable
        logger.debug(request.session.items())
        username = request.session['username']
        token = request.session['token']
        api_server = request.session['api_server']
        emulate = request.session.get('emulate', None)
        return {'username': username, 'token': token, 'api_server': api_server,
                'emulate': emulate}
    except KeyError:
        pass
    try:
        # Attempt #2 - Header/META values, this is DEPRECATED as of v2!
        logger.debug(request.META.items())
        username = request.META['HTTP_X_AUTH_USER']
        token = request.META['HTTP_X_AUTH_TOKEN']
        api_server = request.META['HTTP_X_API_SERVER']
        emulate = request.META.get('HTTP_X_AUTH_EMULATE', None)
        return {'username': username, 'token': token,
                'api_server': api_server, 'emulate': emulate}
    except KeyError:
        pass
    try:
        # Final attempt - GET/POST values
        params = getRequestParams(request)
        logger.debug(params.items())
        username = params['HTTP_X_AUTH_USER']
        token = params['HTTP_X_AUTH_TOKEN']
        api_server = params['HTTP_X_API_SERVER']
        emulate = params.get('HTTP_X_AUTH_EMULATE', None)
        return {'username': username, 'token': token,
                'api_server': api_server, 'emulate': emulate}
    except KeyError:
        pass
    return None


class TokenAuthentication(BaseAuthentication):

    """
    Atmosphere 'AuthToken' based authentication.
    To authenticate, pass the token key in the "Authorization"
    HTTP header, prepended with the string "Token ". For example:
        Authorization: Token 098f6bcd4621d373cade4e832627b4f6
    """
    model = AuthToken

    def authenticate(self, request):
        token_key = None
        auth = request.META.get('HTTP_AUTHORIZATION', '').split()
        if len(auth) == 2 and auth[0].lower() == "token":
            token_key = auth[1]

        if not token_key and 'token' in request.session:
            token_key = request.session['token']
        if validate_token(token_key):
            token = self.model.objects.get(key=token_key)
            if token.user.is_active:
                return (token.user, token)
        return None


class JWTTokenAuthentication(TokenAuthentication):

    """
    JWTTokenAuthentication:
    To authenticate, pass the token key in the "Authorization" HTTP header,
    prepend with the string "Bearer ". For example:
        Authorization: Bearer 098f6bcd4621d373cade4e832627b4f6
    """

    def authenticate(self, request):
        auth = request.META.get('HTTP_AUTHORIZATION', '').split()
        jwt_assertion = request.META.get('HTTP_ASSERTION')
        if jwt_assertion:
            sp = WSO2_JWT(auth_settings.JWT_SP_PUBLIC_KEY_FILE)
            auth_token = sp.create_token_from_jwt(jwt_assertion)
            if auth_token.user.is_active:
                return (auth_token.user, auth_token)
        return None


class GlobusOAuthTokenAuthentication(TokenAuthentication):

    """
    GlobusOAuthTokenAuthentication:
    To authenticate, pass the token key in the "Authorization" HTTP header,
    prepend with the string "Token ". For example:
        Authorization: Token <777-char string>
    """

    def authenticate(self, request):
        auth = request.META.get('HTTP_AUTHORIZATION', '').split()
        if len(auth) == 2 and auth[0].lower() == "token":
            oauth_token = auth[1]
            if validate_globus_oauth_token(oauth_token):
                try:
                    token = self.model.objects.get(key=oauth_token)
                except self.model.DoesNotExist:
                    return None
                if token and token.user.is_active:
                    return (token.user, token)
        return None

def validate_globus_oauth_token(token, request=None):
    """
    Validates the token attached to the request (SessionStorage, GET/POST)
    On every request, ask OAuth to authorize the token
    """
    # Attempt to contact globus
    try:
        user_profile = globus_profile_for_token(token)
    except Exception:
        logger.exception("Globus could not find profile information for token %s" % token)
        user_profile = None

    if not user_profile:
        return False
    # Attempt to 'read' the user_profile
    try:
        auth_token = create_user_token_from_globus_profile(user_profile, token)
    except Exception:
        logger.exception("The method for which to 'read' a globus token has changed. Check the code for more details")
        auth_token = None
    if not auth_token:
        return False
    return True


class OpenstackTokenAuthentication(TokenAuthentication):

    """
    OpenstackTokenAuthentication:
    To authenticate, pass the token key in the "Authorization" HTTP header,
    prepend with the string "Token ". For example:
        Authorization: Token <Keystone_Token>
    """

    def authenticate(self, request):
        """
        TODO: This method might take an already-logged in user who possesses a KEYSTONE TOKEN -- one could then determine if the token was still 'valid' and use that truth-value to authenticate the request.
        The entire user-object would need to be returned, and the dependencies for this method would have to include rtwo/openstacksdk which might make this entire process out-of-scope for django-cyverse-auth
        TODO: Ideally, more things would be passed through headers to tell us:
            - What KEYSTONE_SERVER to authenticate with
            - What the username or other information is *expected* to be..
            - ??
        """
        auth = request.META.get('HTTP_AUTHORIZATION', '').split()
        auth_url = auth_settings.KEYSTONE_SERVER
        region_name = "RegionOne"  # This could be passed in via header _OR_ as an auth_settings.KEYSTONE_REGION_NAME
        domain_name = "Default"  # This could be passed in via header _OR_ as an auth_settings.KEYSTONE_DOMAIN_NAME
        parsed_auth_url = urlparse(auth_url)
        hostname = parsed_auth_url.hostname
        token_key = None
        if len(auth) == 2 and auth[0].lower() == "token":
            token_key = auth[1]
        if not token_key:
            return None
        try:
            from rtwo.drivers.common import _connect_to_openstack_sdk
        except ImportError:
            logger.exception(
                "Cannot use OpenstackTokenAuthentication without `rtwo`."
                " Please `pip install rtwo` and try again!")
            return None

        sdk_args = {
            'auth_url': auth_url.replace('5000', '35357'),
            'ex_force_base_url': auth_url.replace(":5000/v3", ":8774/v2/"),
            'identity_api_version': 3,
            'project_domain_name': domain_name,
            'region_name': region_name,
            'user_domain_name': domain_name,
            "auth_plugin": "token",
            "token": token_key
        }
        stack_sdk = _connect_to_openstack_sdk(**sdk_args)
        try:
            stack_sdk.authorize()
            whoami = stack_sdk.session.auth.auth_ref
            username = whoami.username
            new_profile = {
                'username': username,
                'firstName': username,
                'lastName': "",
                'email': "%s@%s" % (username, hostname),
            }
            logger.debug("Openstack Profile: %s", new_profile)
            user = get_or_create_user(new_profile['username'], new_profile)
            auth_token = get_or_create_token(
                user, token_key, issuer="OpenstackTokenAuthentication")
            return auth_token
        except:
            return None


class OAuthTokenAuthentication(TokenAuthentication):

    """
    OAuthTokenAuthentication:
    To authenticate, pass the token key in the "Authorization" HTTP header,
    prepend with the string "Token ". For example:
        Authorization: Token 098f6bcd4621d373cade4e832627b4f6
    """

    def _mock_oauth_login(self, oauth_token):
        username = settings.ALWAYS_AUTH_USER
        user = get_or_create_user(username, {
            'firstName': "Mocky Mock",
            'lastName': "MockDoodle",
            'email': '%s@iplantcollaborative.org' % settings.ALWAYS_AUTH_USER,
            })
        _, token = self.model.objects.get_or_create(key=oauth_token, user=user)
        return user, token

    def authenticate(self, request):
        all_backends = settings.AUTHENTICATION_BACKENDS
        auth = request.META.get('HTTP_AUTHORIZATION', '').split()
        if len(auth) == 2 and auth[0].lower() == "token":
            oauth_token = auth[1]
            if 'django_cyverse_auth.authBackends.MockLoginBackend' in all_backends:
                user, token = self._mock_oauth_login(oauth_token)
                user_logged_in.send(
                    sender=user.__class__, request=request, user=user)
                return (user, token)
            if validate_oauth_token(oauth_token):
                try:
                    token = self.model.objects.get(key=oauth_token)
                except self.model.DoesNotExist:
                    return None
                if token and token.user.is_active:
                    user = token.user
                    user_logged_in.send(
                        sender=user.__class__, request=request, user=user)
                    return (user, token)
        return None


def validate_oauth_token(token, request=None):
    """
    Validates the token attached to the request (SessionStorage, GET/POST)
    On every request, ask OAuth to authorize the token
    """
    # Attempt to contact CAS
    try:
        user_profile = cas_profile_for_token(token)
    except ConnectionError:
        logger.exception("CAS could not be reached!")
        user_profile = None

    if not user_profile:
        return False
    username = user_profile.get("id")
    if not username:
        logger.warn("Invalid Profile:%s does not have username/attributes"
                    % user_profile)
        return False

    # CAS specific:
    # converts [{u'lastName': u'Doe'}, {u'firstName': u'John'}]
    # to {u'lastName': u'doe', u'firstName': u'John'}
    profile_attrs = {
        c.keys()[0]: c.values()[0]
        for c in user_profile.get('attributes', [])
    }
    username = username.lower()
    new_profile = {
        'username': username,
        'firstName': profile_attrs['firstName'],
        'lastName': profile_attrs['lastName'],
        'email': profile_attrs['email']
    }
    user = get_or_create_user(new_profile['username'], new_profile)
    auth_token = get_or_create_token(
        user, token, issuer="OAuthTokenAuthentication")
    return auth_token


def validate_token(token):
    if not token:
        return False
    try:
        auth_token = AuthToken.objects.get(key=token)
        user = auth_token.user
    except AuthToken.DoesNotExist:
        all_backends = settings.AUTHENTICATION_BACKENDS
        if 'django_cyverse_auth.authBackends.MockLoginBackend' in all_backends:
            logger.info(
                "IGNORED -- AuthToken Retrieved:%s Does not exist. "
                "-- Validate anyway (Mock enabled)" % (token,))
            mock_user, _ = User.objects.get_or_create(
                username=settings.ALWAYS_AUTH_USER)
            auth_token, _ = AuthToken.objects.get_or_create(key=token, user=mock_user)
            return True
        logger.info("AuthToken Retrieved:%s Does not exist." % (token,))
        return False
    if auth_token.is_expired():
	logger.info("Token %s expired, User %s "
		    "could not be reauthenticated in CAS"
		    % (token, user))
	return False
    else:
        return True



"""
Authentication Backends and validation methods
"""
from urlparse import urlparse

from django.contrib.auth.backends import ModelBackend
from django.conf import settings
from django.contrib.auth import get_user_model

from libcloud.common.openstack_identity import OpenStackIdentity_3_0_Connection, OpenStackIdentityTokenScope
from libcloud.compute.providers import get_driver
from libcloud.compute.types import Provider

from keystoneauth1.identity import v3
from keystoneauth1 import session
from keystoneclient.v3 import client


from .settings import auth_settings
from .models import get_or_create_user, get_or_create_token
from .models import Token
from .protocol.ldap import ldap_validate, ldap_formatAttrs
from .protocol.ldap import lookupUser as ldap_lookupUser
from .protocol.cas import cas_validateUser
from .protocol.globus import (
    globus_validate_code, _extract_first_last_name,
    globus_profile_for_token,
    create_user_token_from_globus_profile)
from caslib import OAuthClient as CAS_OAuthClient
#From troposphere
import ldap
import logging

logger = logging.getLogger(__name__)


from rest_framework import authentication, exceptions
from uuid import uuid4


class SAMLLoginBackend(ModelBackend):

    """
    Implemting an AuthenticationBackend
    (Used by Django for logging in to admin, storing session info)
    """

    def authenticate(self, request, username=None, password=None):
        """
        Return user if validated by CAS
        Return None otherwise.
        """
        # logger.debug("SAMLBackend-- U:%s P:%s R:%s"
        #              % (username, password, request))
        if not request:
            logger.debug("SAML Authentication skipped - No request.")
            return None
        # TODO: See if you were the auth backend used to originate the request.
        # TODO: Look at request session for a token and see if its still valid.
        if False:
            logger.debug("SAML Authentication failed - %s" % username)
            return None


class CASLoginBackend(ModelBackend):

    """
    Implemting an AuthenticationBackend
    (Used by Django for logging in to admin, storing session info)
    """

    def authenticate(self, request, username=None, password=None):
        """
        Return user if validated by CAS
        Return None otherwise.
        """
        # logger.debug("CASBackend -- U:%s P:%s R:%s"
        #              % (username, password, request))
        if not username:
            logger.debug("CAS Authentication skipped - No Username.")
            return None
        (success, cas_response) = cas_validateUser(username)
        logger.info("Authenticate by CAS: %s - %s %s"
                    % (username, success, cas_response))
        if not success:
            logger.debug("CAS Authentication failed - %s" % username)
            return None
        attributes = cas_response.attributes
        return get_or_create_user(username, attributes)


class LDAPLoginBackend(ModelBackend):

    """
    AuthenticationBackend for LDAP logins
    (Logging in from admin or Django REST framework login)
    """

    def authenticate(self, request, username=None, password=None, token=None):
        """
        Return user if validated by LDAP.
        Return None otherwise.
        """
        # logger.debug("LDAPBackend-- U:%s P:%s R:%s"
        #              % (username, password, request))
        if not ldap_validate(username, password):
            logger.debug("LDAP Authentication failed - %s" % username)
            return None
        ldap_attrs = ldap_lookupUser(username)
        attributes = ldap_formatAttrs(ldap_attrs)
        attributes['username'] = username
        logger.debug("[LDAP] Authentication Success - " + username)
        return self._update_token(attributes, token, request)

    def _update_token(self, user_profile, token, request=None):
        user = get_or_create_user(user_profile['username'], user_profile)
        auth_token = get_or_create_token(user, token, issuer="OpenstackLoginBackend")
        user = auth_token.user
        if request:
            request.session['token_key'] = auth_token.key
        return user



class AuthTokenLoginBackend(ModelBackend):

    """
    AuthenticationBackend for OAuth authorizations
    (Authorize user from Third party (web) clients via OAuth)
    """
    def __init__(self, *args, **kwargs):
        super(AuthTokenLoginBackend, self).__init__(*args, **kwargs)

    def authenticate(self, request, username=None, password=None, auth_token=None):
        """
        Return user if validated by their auth_token
        Return None otherwise.
        """
        try:
            valid_token = Token.objects.get(key=auth_token)
        except Token.DoesNotExist:
            return None
        if valid_token.is_expired():
            logger.debug(
                "[AUTHTOKEN] Token %s is expired. (User:%s)"
                % (valid_token.key, valid_token.user))
            return None
        logger.debug(
            "[AUTHTOKEN] Valid Token %s (User:%s)"
            % (valid_token.key, valid_token.user))
        valid_user = valid_token.user
        return get_or_create_user(valid_user.username, {})

#### Troposphere needs

cas_oauth_client = CAS_OAuthClient(auth_settings.CAS_SERVER,
                                   auth_settings.OAUTH_CLIENT_CALLBACK,
                                   auth_settings.OAUTH_CLIENT_KEY,
                                   auth_settings.OAUTH_CLIENT_SECRET,
                                   auth_prefix=auth_settings.CAS_AUTH_PREFIX)


def create_user_token_from_cas_profile(profile, access_token):
    profile_dict = dict()
    username = profile['id']
    for attr in profile['attributes']:
        key = attr.keys()[0]
        value = attr[key]
        profile_dict[key] = value

    user = get_or_create_user(username, profile_dict)
    user_token = Token.objects.create(key=access_token, user=user)
    return user_token

class GlobusOAuthLoginBackend(object):
    """
    Globus OAuth Authentication Backend

    Exchanges an access_token for a user, creates if does not exist
    """

    def authenticate(self, request, key=None):
        user_token = None
        try:
            user_token = Token.objects.get(key=key)
        except Token.DoesNotExist:
            user_profile = globus_profile_for_token(key)
            user_token = create_user_token_from_globus_profile(user_profile, key)
        if not user_token:
            return None
        user = user_token.user
        return user

    def get_user(self, user_id):
        """
        Get a User object from the username.
        """
        User = get_user_model()
        try:
            return User.objects.get(pk=user_id)
        except User.DoesNotExist:
            return None


class OAuthLoginBackend(object):
    """
    CAS OAuth Authentication Backend

    Exchanges an access_token for a user, creates if does not exist
    """

    def authenticate(self, request, access_token=None):
        try:
            user_token = Token.objects.get(key=access_token)

        except Token.DoesNotExist:
            try:
                profile = cas_oauth_client.get_profile(access_token=access_token)
            except:
                logger.exception("Unexpected error occurred during exchange of access_token --> profile")
                raise exceptions.AuthenticationFailed("Unexpected error occurred during exchange of access_token --> profile")
            error = profile.get('error')

            if error:
                raise exceptions.AuthenticationFailed(error)
            user_token = create_user_token_from_cas_profile(profile, access_token)

        user = user_token.user
        return user

    def get_user(self, user_id):
        """
        Get a User object from the username.
        """
        User = get_user_model()
        try:
            return User.objects.get(pk=user_id)
        except User.DoesNotExist:
            return None


class OAuthTokenLoginBackend(authentication.BaseAuthentication):
    """
    CAS OAuth Authentication Backend

    Exchanges an access_token for a user, creates if does not exist
    """

    def authenticate(self, request):
        access_token = None
        auth = request.META.get('HTTP_AUTHORIZATION', '').split()
        if len(auth) == 2 and auth[0].lower() == "token":
            access_token = auth[1]
        else:
            return None

        try:
            user_token = Token.objects.get(key=access_token)

        except Token.DoesNotExist:
            try:
                profile = cas_oauth_client.get_profile(access_token=access_token)
            except:
                logger.exception("Unexpected error occurred during exchange of access_token --> profile")
                raise exceptions.AuthenticationFailed("Unexpected error occurred during exchange of access_token --> profile")
            error = profile.get('error')

            if error:
                raise exceptions.AuthenticationFailed(error)

            user_token = create_user_token_from_cas_profile(profile, access_token)

        user = user_token.user
        return (user, user_token)


class MockLoginBackend(authentication.BaseAuthentication):
    """
    AuthenticationBackend for Testing login
    (Logging in from admin or Django REST framework login)
    """
    def authenticate(self, request, username=None, password=None):
        """
        Return user if Always
        Return None Never.
        """
        return get_or_create_user(settings.ALWAYS_AUTH_USER, {
            'username':settings.ALWAYS_AUTH_USER,
            'firstName':"Mocky Mock",
            'lastName':"MockDoodle",
            'email': '%s@iplantcollaborative.org' % settings.ALWAYS_AUTH_USER,
            'entitlement': []
        })

    def get_user(self, user_id):
        """
        Get a User object from the username.
        """
        User = get_user_model()
        try:
            return User.objects.get(pk=user_id)
        except User.DoesNotExist:
            return None


class OpenstackLoginBackend(ModelBackend):
    """
    Keystone Login for Atmosphere
    Includes optional libcloud setup if/when keystone client stops working.
    """
    strategy = "keystone"

    def authenticate(self, request, username, password, project_name=None, auth_url=None, token=None, domain=None):
        if not project_name:
            project_name = username

        if not auth_url:
            auth_url = auth_settings.KEYSTONE_SERVER
        if not domain:
            domain = auth_settings.KEYSTONE_DOMAIN_NAME
        if '/v' not in auth_url:
            auth_url += "/v3"  # Assume v3
        #TODO: If necessary, create a auth_setting 'feature' to select 'libcloud' or 'keystone' as the strategy and then validate token/auth the same way.
        if self.strategy == 'libcloud':
            user = self.auth_by_libcloud(auth_url, project_name, domain, username, password, token, request)
        else:
            user = self.auth_by_keystone(auth_url, project_name, domain, username, password, token, request)
        return user

    def auth_by_keystone(self, auth_url, project_name, domain, username, password=None, token=None, request=None):
        """
        Given username/password or username/token
        return user
        """
        if token:
            user = self.keystone_validate_token(auth_url, username, token, project_name, domain, request)
            return user
        return self.keystone_validate_auth(auth_url, username, password, project_name, domain, request)

    def keystone_validate_auth(self, auth_url, username, password, project_name, domain_name, request):
        """
        Given username,password -- validate with keystone
        """
        password_auth = v3.Password(
            auth_url=auth_url,
            username=username, password=password,
            user_domain_id=domain_name,
            project_name=project_name, project_domain_id=domain_name)

        try:
            ks_session = session.Session(auth=password_auth)
            ks_session.get_token()
        except:
            logger.exception("Error: Keystone session did not produce a valid token")
            return None

        try:
            expiry_time = password_auth.auth_ref.expires
            email = self._lookup_email(ks_session)
            _, token = self._keystone_auth_to_token(password_auth, username, project_name)
            return self._update_token(auth_url, username, token, email, expiry_time, request)
        except:
            logger.exception("Error parsing keystone auth by password")
            return None

    def keystone_validate_token(self, auth_url, username, token, project_name, project_domain, request):
        """
        Given token -- validate with keystone
        """
        token_auth=v3.Token(
            auth_url=auth_url,
            token=token,
            project_name=project_name,
            project_domain_id=project_domain)

        ks_session=session.Session(auth=token_auth)
        auth_token = ks_session.get_token()
        expiry_time = token_auth.auth_ref.expires
        try:
            email = self._lookup_email(ks_session)
            self._keystone_auth_to_token(token_auth, username, project_name)
            return self._update_token(auth_url, username, token, email, expiry_time, request)
        except:
            logger.exception("Error validating keystone auth by token")
            return None

    def _keystone_auth_to_token(self, keystone_auth_obj, username, project_name):
        """
        Validates a keystone auth (v3.Password or v3.Token) to determine if its valid
        if valid, it should match the username/project_name
        """
        ks_session = session.Session(auth=keystone_auth_obj)
        ks_client = client.Client(session=ks_session)
        # Validate the driver by requesting token data from keystone
        try:
            token_key = ks_session.get_token()
        except:
            raise Exception("Keystone client could not validate authentication")

        # Validate the token_data returned from keystone
        try:
            token_data = ks_client.tokens.get_token_data(token_key)
            token_username = token_data['token']['user']['name']
            token_project_name = token_data['token']['project']['name']
        except (KeyError, ValueError):
            logger.exception("DATA CHANGED -- update value to match new token_data: %s" % token_data)
            raise

        if token_project_name != project_name:
            raise Exception("Token %s does not match expected project name - %s" % token_key, project_name)
        if token_username != username:
            raise Exception("Token %s does not match expected username - %s" % token_key, username)
        return ks_session, token_key
    #Alternative method -- libcloud 'strategy'

    def auth_by_libcloud(self, auth_url, project_name, domain, username, password=None, token=None, request=None):
        """
        Given username/password or username/token
        return user
        """
        if token:
            user = self.libcloud_validate_token(auth_url, username, token, project_name, domain, request)
            return user
        return self.libcloud_validate_auth(auth_url, username, password, project_name, domain, request)

    def libcloud_validate_token(self, auth_url, username, token, project_name, domain, request):
        OpenStack = get_driver(Provider.OPENSTACK)
        driver = OpenStack(username, "",
                   ex_force_base_url=auth_url.replace(":5000/v3",":8774/v2"),
                   ex_force_auth_url=auth_url,
                   ex_tenant_name=project_name, ex_domain_name=domain,
                   ex_force_auth_version='3.x_password',
                   ex_force_service_region='RegionOne',
                   ex_force_auth_token=token)
        try:
            sizes = driver.list_sizes()
            return self._update_token(auth_url, username, token, request)
        except Exception as exc:
            logger.exception("Error validating libcloud auth by token")
            return None


    def libcloud_validate_auth(self, auth_url, username, password, project_name, domain, request):
        driver = OpenStackIdentity_3_0_Connection(
            auth_url=auth_url+"/auth/tokens",
            user_id=username, key=password,  #TODO: add domain
            token_scope=OpenStackIdentityTokenScope.PROJECT, tenant_name=project_name)

        try:
            conn = driver.authenticate()
            auth_token = driver.auth_token
            return self._update_token(auth_url, username, auth_token, request)
        except:
            logger.exception("Error validating libcloud auth by password")
            return None

    # Private helper methods -- commonly used

    def _user_profile_for_auth(self, auth_url, username, email):
        parsed_auth_url = urlparse(auth_url)
        hostname = parsed_auth_url.hostname
        user_profile = {
            'username': username,
            'firstName': username,
            'lastName': "",
            'email': email,
            'entitlement': []
        }
        return user_profile

    def _update_token(self, auth_url, username, token, email, expiry_time, request=None):
        user_profile = self._user_profile_for_auth(auth_url, username, email)
        user = get_or_create_user(user_profile['username'], user_profile)
        auth_token = get_or_create_token(user, token, token_expire=expiry_time, issuer="OpenstackLoginBackend")
        user = auth_token.user
        if request:
            request.session['token_key'] = auth_token.key
            request.session['token_expire_time'] = str(expiry_time)[:19]
        return user

    def _grant_access(self, auth_url, username):
        user_profile = self._user_profile_for_auth(auth_url, username)
        return get_or_create_user(username, user_profile)

    def _lookup_email(self, session):
        try:
            user_id = str(session.get_user_id())
            ks_client = client.Client(session=session)
            ks_user = ks_client.users.get(user_id)
            email = str(ks_user.email)
        except keystoneclient.exceptions:
            #Still thinking about how to obtain the error message
            return None
        return email



"""
LDAP authentication methods
"""
from __future__ import absolute_import
from django_cyverse_auth.settings import auth_settings
import string

from django.core.handlers.wsgi import WSGIRequest
from django.utils import timezone
import ldap as ldap_driver

import logging
logger = logging.getLogger(__name__)



def _get_uid_number(userid):
    """
    Get uidNumber
    """
    try:
        conn = ldap_driver.initialize(auth_settings.LDAP_SERVER)
        attr = conn.search_s(auth_settings.LDAP_SERVER_DN,
                             ldap_driver.SCOPE_SUBTREE,
                             "(uid=%s)" % userid)
        uid_number = int(attr[0][1]["uidNumber"][0])
        if uid_number > 10000:
            uid_number -= 10000
        return uid_number
    except IndexError:
        logger.warn("Error - User %s does not exist" % userid)
        return None
    except Exception as e:
        logger.warn(
            "Error occurred getting user uidNumber for user: %s" %
            userid)
        logger.exception(e)
        return None

def _search_ldap(userid, conn=None):
    try:
        if not conn:
            conn = ldap_driver.initialize(auth_settings.LDAP_SERVER)
        result = conn.search_s(
            auth_settings.LDAP_SERVER_DN,
            ldap_driver.SCOPE_SUBTREE,
            '(uid=' + userid + ')'
        )
        return result
    except Exception as e:
        logger.warn("Error occurred on ldap search for: %s" % userid)
        logger.exception(e)
        return None


def getAllUsers():
    """
    Grabs all users in LDAP
    """
    try:
        conn = ldap_driver.initialize(auth_settings.LDAP_SERVER)
        user_list = []
        for letter in string.lowercase:
            attr = _search_ldap("%s*" % letter, conn)
            for i in xrange(0, len(attr)):
                user_attrs = attr[i][1]
                user_list.append(user_attrs)
        return user_list
    except Exception as e:
        logger.warn("Error occurred looking up all user")
        logger.exception(e)
        return None


def user_expiry_stats(profile):
    epoch = timezone.datetime(1970, 1, 1)
    try:
        username = profile['uid'][0]
        change = profile['shadowLastChange'][0]
        max_val = profile['shadowMax'][0]
        warn = profile['shadowWarning'][0]

        chgdate = epoch + timezone.timedelta(days=int(change))
        expiry_date = chgdate+timezone.timedelta(days=int(max_val))
        warndate = expiry_date+timezone.timedelta(days=-int(warn))

        if chgdate == epoch:
            logger.warn("Could not calculate an expiration for user %s" % username)
    except (KeyError, IndexError):
        logger.warn("Could not calculate an expiration for profile %s" % profile)
        chgdate = epoch
        warndate = epoch
        expiry_date = timezone.datetime.max
    return {
        'last_changed': chgdate,
        'expires_on': expiry_date,
        'warn_user_on': warndate
    }


def lookupUser(userid):
    """
    Grabs email for the user based on LDAP attrs
    """
    try:
        results = _search_ldap(userid)
        user_dn, user_attrs = results[0]
        expiry_dict = user_expiry_stats(user_attrs)
        user_attrs.update({'expiry': expiry_dict})
        return user_attrs
    except Exception as e:
        logger.warn("Error occurred looking up user: %s" % userid)
        logger.exception(e)
        raise


def lookupEmail(userid):
    """
    Grabs email for the user based on LDAP attrs
    """
    try:
        logger.debug(type(userid))
        if isinstance(userid, WSGIRequest):
            raise Exception("WSGIRequest invalid.")
        attr = _search_ldap(userid)
        emailaddr = attr[0][1]['mail'][0]
        return emailaddr
    except Exception as e:
        logger.warn("Error occurred looking up email for user: %s" % userid)
        logger.exception(e)
        raise


def ldap_validate(username, password):
    """
    ldap_validate
    Using the username and password parameters, test with an LDAP bind.
    If the connection succeeds, the credentials are authentic.
    """
    if not username or not password:
        logger.warn("[LDAP] Skip Test - Username/Password combination missing")
        return

    try:
        ldap_server = auth_settings.LDAP_SERVER
        ldap_server_dn = auth_settings.LDAP_SERVER_DN
        logger.warn("[LDAP] Validation Test - %s" % username)
        ldap_conn = ldap_driver.initialize(ldap_server)
        dn = "uid=" + username + "," + ldap_server_dn
        ldap_conn.simple_bind_s(dn, password)
        return True
    except Exception as e:
        logger.exception(e)
        return False


def ldap_formatAttrs(ldap_attrs):
    """
    Formats attrs into a unified dict to ease in user creation
    """
    logger.info(ldap_attrs)
    try:
        return {
            'email': ldap_attrs['mail'][0],
            'firstName': ldap_attrs['givenName'][0],
            'lastName': ldap_attrs['sn'][0],
        }
    except KeyError as nokey:
        logger.exception(nokey)
        return None


def get_groups_for(username):
    try:
        ldap_server = auth_settings.LDAP_SERVER
        ldap_group_dn = auth_settings.LDAP_SERVER_DN.replace(
            "ou=people", "ou=Groups")
        ldap_conn = ldap_driver.initialize(ldap_server)
        matching_groups = ldap_conn.search_s(
            ldap_group_dn, ldap_driver.SCOPE_SUBTREE, "(&(cn=*)(memberUid=%s))" % username)
        groups = []
        for group in matching_groups:
            group_name = group[1]['cn'][0]
            groups.append(group_name)
        return groups
    except Exception as e:
        logger.exception(e)
        return []


def list_groups():
    groups = []
    try:
        ldap_server = auth_settings.LDAP_SERVER
        ldap_group_dn = auth_settings.LDAP_SERVER_DN.replace(
            "ou=people", "ou=Groups")
        ldap_conn = ldap_driver.initialize(ldap_server)
        matching_groups = ldap_conn.search_s(
            ldap_group_dn, ldap_driver.SCOPE_SUBTREE, "(&(cn=*))")
        for group in matching_groups:
            group_name = group[1]['cn'][0]
            groups.append(group_name)
    except Exception as e:
        logger.exception(e)
    return groups

def get_members(groupname):
    """
    """
    try:
        ldap_server = auth_settings.LDAP_SERVER
        ldap_group_dn = auth_settings.LDAP_SERVER_DN.replace(
            "ou=people", "ou=Groups")
        ldap_conn = ldap_driver.initialize(ldap_server)
        group_users = ldap_conn.search_s(
            ldap_group_dn, ldap_driver.SCOPE_SUBTREE, '(cn=%s)' % groupname)
        return group_users[0][1]['memberUid']
    except Exception as e:
        logger.exception(e)
        return []


def is_staff(username):
    """
    ldap_validate
    Using the username is in the atmo-user group return True
    otherwise False.
    """
    return is_user_in_group(username, 'staff')


def is_atmo_user(username):
    """
    ldap_validate
    Using the username is in the atmo-user group return True
    otherwise False.
    """
    return is_user_in_group(username, 'atmo-user')


def is_user_in_group(username, groupname):
    members_list = get_members(groupname)
    return username in members_list


def get_atmo_users():
    """
    """
    members_list = get_members('atmo-user')
    return members_list


def get_core_services():
    """
    """
    members_list = get_members('core-services')
    return members_list


def get_staff_users():
    """
    """
    members_list = get_members('staff')
    return members_list



import requests
import json, os

from base64 import b64encode
from django.http import HttpResponse, HttpResponseRedirect
from django.utils import timezone

from oauth2client.client import OAuth2WebServerFlow
from oauth2client.client import Error as OAuthError

from django_cyverse_auth.models import (
    get_or_create_user, get_or_create_token,
    create_access_token, get_access_token)
from django_cyverse_auth.settings import auth_settings
from django_cyverse_auth.exceptions import Unauthorized

import logging
logger = logging.getLogger(__name__)


def globus_initFlow():
    """
    Retrieve cached/Create a new access token
    and use it to create an OAuth2WebServerFlow
    """
    # Use client_id:client_secret for authorization
    userAndPass = "%s:%s" % (auth_settings.GLOBUS_OAUTH_ID, auth_settings.GLOBUS_OAUTH_SECRET)
    b64_userAndPass = b64encode(userAndPass)
    auth_header = "Basic %s" % b64_userAndPass
    flow = OAuth2WebServerFlow(
        client_id=auth_settings.GLOBUS_OAUTH_ID,
        scope=auth_settings.GLOBUS_OAUTH_ATMOSPHERE_SCOPE,
        authorization_header=auth_header,
        redirect_uri=auth_settings.OAUTH_CLIENT_CALLBACK,
        auth_uri=auth_settings.GLOBUS_AUTH_URL,
        token_info_uri=auth_settings.GLOBUS_TOKENINFO_URL,
        token_uri=auth_settings.GLOBUS_TOKEN_URL)
    return flow

def globus_logout(redirect_uri, redirect_name='Jetstream'):
    """
    Redirect to logout of globus
    """
    flow = globus_initFlow()
    auth_uri = flow.auth_uri
    web_logout_url = auth_uri.replace('oauth2/authorize', 'web/logout')
    web_logout_url += "?client_id=%s&redirect_name=%s&redirect_uri=%s"\
            % (flow.client_id, redirect_name, redirect_uri)
    logger.info(web_logout_url)
    return HttpResponseRedirect(web_logout_url)

def globus_authorize(request):
    """
    Redirect to the IdP based on 'flow'
    """
    flow = globus_initFlow()
    auth_uri = flow.step1_get_authorize_url()
    auth_uri += '&authentication_hint=36007761-2cf2-4e74-a068-7473afc1d054'
    auth_uri = auth_uri.replace('access_type=offline','access_type=online')
    logger.warn(auth_uri)
    return HttpResponseRedirect(auth_uri)

def globus_profile_for_token(globus_user_token):
    try:
        logger.info("Request Token Info for key %s" % globus_user_token)
        userAndPass = "%s:%s" % (auth_settings.GLOBUS_OAUTH_ID, auth_settings.GLOBUS_OAUTH_SECRET)
        b64_userAndPass = b64encode(userAndPass)
        auth_header = "Basic %s" % b64_userAndPass
        r = requests.post(
            auth_settings.GLOBUS_TOKENINFO_URL+'?include=effective', verify=False,
            data={'token':globus_user_token},
            headers={'Authorization':auth_header})
        j_data = r.json()
        logger.info(j_data)
        return j_data
    except:
        logger.exception("Error retrieving profile from globus")
        return None

def _extract_expiry_date(epoch_secs):
    if type(epoch_secs) != int:
        try:
            epoch_secs = int(date_int)
        except:
            logger.exception("Expected date to be an integer (Seconds from epoch). This method should be modified to the new use case for globus 'exp'")
            return None
    epoch = timezone.datetime(month=1, day=1, year=1970)
    return epoch + timezone.timedelta(seconds=epoch_secs)

def _extract_first_last_name(user_name):
    if ' ' not in user_name:
        return '', user_name
    split_name = user_name.split()
    return split_name[0], ' '.join(split_name[1:])

def _extract_user_from_email(raw_username):
    """
    Usernames come from the globus provider in the form:
    username@login_authority.com
    """
    if not raw_username:
        return None
    return raw_username.split('@')[0]

def _map_email_to_user(raw_username):
    """
    Input:  test@fake.com
    Output: test
    """
    if not auth_settings.GLOBUS_MAPPING_FILE:
        logger.info("GLOBUS_MAPPING_FILE NOT defined. Check your auth settings!!")
        return raw_username
    if not os.path.exists(auth_settings.GLOBUS_MAPPING_FILE):
        logger.warn("GLOBUS_MAPPING_FILE %s does not exist!" % auth_settings.GLOBUS_MAPPING_FILE)
        return None
    try:
        with open(auth_settings.GLOBUS_MAPPING_FILE) as the_file:
            text = the_file.read()
            user_mapping = json.loads(text)
    except:
        logger.warn("GLOBUS_MAPPING_FILE %s is NOT VALID JSON!" % auth_settings.GLOBUS_MAPPING_FILE)
        user_mapping = {}

    if raw_username not in user_mapping:
        return None
    username = user_mapping[raw_username]
    logger.info("GLOBUS_MAPPING_FILE identified %s -> %s" % (raw_username, username))
    return username


def parse_atmosphere_token(token_response):
    atmosphere_scope = auth_settings.get('GLOBUS_OAUTH_ATMOSPHERE_SCOPE', 'urn:globus:auth:scope:use.jetstream-cloud.org:all')
    atmosphere_scope = atmosphere_scope.replace('openid email profile ','').strip()
    tokens = token_response.pop('other_tokens', [])
    tokens.append(token_response)
    for token in tokens:
        #FIXME: Make this a specific setting, rather than hard-coded.
        if token['scope'] == atmosphere_scope:
            return token['access_token']
    raise Exception("Globus token valid -- Could not find a token with atmosphere scope %s" % atmosphere_scope)

def globus_validate_code(request):
    """
    This flow is used to create a new Token on behalf of a Service Client
    (Like Troposphere)
    Validates 'code' returned from the IdP
    If valid: Return new AuthToken to be passed to the Resource Provider.
        else: Return None
    """
    code = request.GET.get('code','')
    error = request.GET.get('error','')
    error_description = request.GET.get('error_description','')
    if error:
        error_msg = "%s: %s" % (error, error_description) if error_description else error
        raise Unauthorized(error_msg)
    if not code:
        logger.warn("User returned from Login prompt but there was NO `code` to validate!")
        return None
    if type(code) == list:
        code = code[0]
    flow = globus_initFlow()
    try:
        credentials = flow.step2_exchange(code)
        logger.info(credentials.__dict__)
    except OAuthError as err:
        logger.exception("Error exchanging code w/ globus")
        return None
    except Exception as err:
        logger.exception("Unknown Error occurred while exchanging code w/ globus")
        return None
    # Parsing
    try:
        user_access_token = parse_atmosphere_token(credentials.token_response)
        token_profile = credentials.id_token
        expiry_date = credentials.token_expiry
    except Exception as err:
        logger.exception("Parse of the credentials response failed. Ask a developer for help!")
        return None
    raw_username = token_profile['preferred_username']
    email = token_profile['email']
    username = _extract_user_from_email(raw_username)
    if not username:
        logger.info("No user provided in token_profile: Check output %s" % token_profile)
        return None
    full_name = token_profile['name']
    issuer = token_profile['iss']
    # Creation
    first_name, last_name = _extract_first_last_name(full_name)
    username = username.lower()
    user_profile = {
        'username':username,
        'firstName':first_name,
        'lastName':last_name,
        'email': email,
    }
    user = get_or_create_user(user_profile['username'], user_profile)
    auth_token = get_or_create_token(user, user_access_token, token_expire=expiry_date, issuer="OpenstackLoginBackend")
    return auth_token

def create_user_token_from_globus_profile(profile, access_token):
    """
    Use this method on your Resource Provider (Like Atmosphere)
    to exchange a profile (that was retrieved via a tokeninfo endpoint)
    for a UserToken that can then be internally validated in an 'authorize' authBackend step..
    """

    logger.info(profile)
    expiry = profile['exp'] # This is an 'epoch-int'
    expiry = _extract_expiry_date(expiry)
    issuer = profile['iss']
    issued_at = profile['iat']
    raw_username = profile['username']  # username@login_auth.com
    raw_name = profile['name']
    email = profile['email']
    username = _extract_user_from_email(raw_username)
    first_name, last_name = _extract_first_last_name(raw_name)
    profile_dict = {
        'username':username,
        'firstName':first_name,
        'lastName':last_name,
        'email': email,
    }
    user = get_or_create_user(profile_dict['username'], profile_dict)
    auth_token = get_or_create_token(user, access_token, token_expire=expiry, issuer=issuer)
    return auth_token
